from scipy import stats
import numpy as np
import cv2
import math
import time
import glob, os
import random
import ctypes
from datetime import datetime
import skimage
import skimage.io
import fnmatch
from skimage.color import rgb2gray
from util.flow_utils import single_warp_noTorch, single_warp_Torch

import torchvision.transforms as transforms
from collections import OrderedDict
import json

from matplotlib import pyplot as plt
plt.switch_backend('agg')


def plot_test(data, label2plot = 'PSNR', savefolder='./results'):
    plt.figure(1)
    lw = 0.5
    for video in data.keys():
        ydata = data[video][label2plot]
        xdata = np.array(range(len(ydata)))
        colorvec = np.random.rand(3)
        plt.plot(xdata, ydata, color=colorvec, lw=lw, label=video)    
    # plt.xlim([0, 100])
    plt.ylim([0.0, 50])
    plt.xlabel('Frames')
    plt.ylabel('PSNR')
    plt.title('PSNR curves')
    plt.legend(loc="lower right")
    plt.savefig(savefolder+'/PSNRcurves.png', format='png', dpi=300)
    plt.close(1)

def print_dict(val_losses, suffix="_valLoss", savefile=None):
    val_losses = dict([(k+suffix, v) for k, v in val_losses.items()])
    message = "["
    for k, v in val_losses.items():
        message += '%s: %.3f, ' % (k, v)
    message = message[:-2] + "]"
    print(message)
    if not savefile is None:
        with open(savefile, "a") as log_file:
            log_file.write('%s\n' % message)  # save the message

def add_loss(vloss, vname, loss):
    if not vname in vloss.keys():
        vloss[vname] = OrderedDict()
        for k, v in loss.items():
            vloss[vname][k] = [v]
    else:
        for k, v in loss.items():
            vloss[vname][k].append(v)
    return vloss

def save_ordereddict(d,filepath):
    with open(filepath, 'w') as f:
        f.write(json.dumps(list(d.items())))

def load_ordereddict(filepath):
    with open(filepath, 'r') as read_file:
        dvec = json.loads(read_file.read())
    dout = OrderedDict()
    for d in dvec:
        for i in range(int(len(d)/2)):
            dout[d[2*i]] = d[2*i+1]
    return dout

def define_transforms(opt):
    """ Defines a transformation (T) and its inverse (iT) to go from
    the dataset to the input of the network.
    iT is used in testing the recurrency or whenever ones needs to go from
    the output of the network back to a numpy image, do some process, and then go back
    to and feed it to the network again.
    """
    if opt.input_nc == 1:
        mean = 0.5*np.ones(opt.patch_depth,dtype=np.float32)
        std = 0.5*np.ones(opt.patch_depth,dtype=np.float32)
    else:
        mean = 0.5*np.ones(opt.patch_depth*opt.output_nc,dtype=np.float32)
        std = 0.5*np.ones(opt.patch_depth*opt.output_nc,dtype=np.float32)

    transform_list = [transforms.ToTensor()]
    if opt.input_nc == 1:  # if grayscale:
        transform_list += [transforms.Grayscale(1), transforms.Normalize(mean, std)]
    else:
        transform_list += [transforms.Normalize(mean, std)]
    T = transforms.Compose(transform_list)
    
    transform_list = [transforms.Normalize((-mean / std).tolist(), (1.0 / std).tolist())]
    transform_list += [transforms.Lambda(lambda x: x.permute(1, 2, 0).numpy())]
    iT = transforms.Compose(transform_list)
    return T, iT

def iio_write(arr,path):
    skimage.io.imsave(path,arr,photometric='rgb', check_contrast=False)

def iio_read(path):
    return np.atleast_3d(skimage.io.imread(path))

def iio_imshow(img):
    img = img.astype(np.uint8)
    c = img.shape[2]
    if c==1:
        r, g, b = img[:, :, 0], img[:, :, 0], img[:, :, 0]
    elif c==4:
        r, g, b = img[:, :, 0], img[:, :, 1], img[:, :, 2]
    else:
        r,g,b = cv2.split(np.uint8(np.array(img)))
    img_bgr = cv2.merge([b,g,r])
    cv2.imshow('img',img_bgr)
    cv2.waitKey(0)


def get_files_pattern(d, pattern):
    """
    List elements in the directory d with pattern.
    Sort the elements.
    """
    files = os.listdir(d)
    files = fnmatch.filter(files, pattern)
    return sorted(files)

def list_video_files_at_dir(d):
    paths = get_files_pattern(d, '*tiff')
    if len(paths) == 0:
        paths = get_files_pattern(d, '*tif')
    if len(paths) == 0:
        paths = get_files_pattern(d, '*png')
    if len(paths) == 0:
        paths = get_files_pattern(d, '*jpg')
    if len(paths) == 0:
        paths = get_files_pattern(d, '*jpeg')
    if len(paths) == 0:
        paths = get_files_pattern(d, '*raw')
    assert (len(paths) > 0), "%s is empty!" % d
    return [os.path.join(d, p) for p in paths]

def load_image(path, ftype='8bits'):
    img = iio_read(path)
    img = np.asarray(img, dtype=np.float32)
    if ftype == '8bits':
        img = img/255.
    elif ftype == '10bits':
        img = img/1023.
    elif ftype == '12bits':
        img = img/4095.
    return img


def pathdiff(a,b):
    assert a[:len(b)]==b, "b should be a subfolder/subfile of a"
    res = os.path.dirname(a[len(b):])
    if res[0]=='/':
        return res[1:]
    else:
        return res

def warpedimagefile(wfolder, fromCode, toCode):
    return os.path.join(wfolder,fromCode+'_'+toCode+'.tif')

class CPPbridge(object):
    def __init__(self,libpath):
        self.libBridge = ctypes.cdll.LoadLibrary(libpath)

        self.libBridge.inverse_compositional.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_double]
        self.libBridge.inverse_compositional.restype = None

        self.libBridge.tvl1flow.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
        self.libBridge.tvl1flow.restype = None

        self.libBridge.blur.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_bool, ctypes.c_bool, ctypes.c_bool]
        self.libBridge.blur.restype = None

        self.libBridge.downsa.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_char]
        self.libBridge.downsa.restype = None
        
        self.libBridge.upsa.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_float, ctypes.c_float]
        self.libBridge.upsa.restype = None

    def invcomp_flow(self, Im1, Im2, nscales=0, nparams=8, robust=2, lambda_value=100.0):   
        """ Computes the optical flow from the Inverse Compositional Algorithm.
            ICA returns a matrix which is translated into optical flow.
            
            So,  
            >>> cv2.warpPerspective(Im1, warp_matrix.reshape(3, 3), (Im1.shape[1],Im1.shape[0]))
            equals
            >>> single_warp_Torch(Im1,flow)
        """   
        h, w = Im1.shape[:2]
        h1, w1 = Im2.shape[:2]
        assert h1==h and w1==w, "Both images Im1 and Im2 are supposed to share same size"

        I1 = np.zeros(h*w, dtype = ctypes.c_double)
        I2 = np.zeros(h*w, dtype = ctypes.c_double)
        flow = np.zeros(2*h*w, dtype = ctypes.c_float)
        warp_matrix = np.zeros(9, dtype = ctypes.c_double)
        
        if len(Im1.shape)==3 and Im1.shape[2]==3:
            I1[:] = rgb2gray(Im1).flatten()[:]
            I2[:] = rgb2gray(Im2).flatten()[:]
        elif len(Im1.shape) == 3 and Im1.shape[2] == 4:
            I1[:] = np.mean(Im1, axis=2).flatten()[:]
            I2[:] = np.mean(Im2, axis=2).flatten()[:]
        elif len(Im1.shape) == 2 or Im1.shape[2] == 1:
            I1[:] = Im1.flatten()[:]
            I2[:] = Im2.flatten()[:]
        
        doublep = ctypes.POINTER(ctypes.c_double)
        self.libBridge.inverse_compositional(I2.ctypes.data_as(doublep), I1.ctypes.data_as(
            doublep), ctypes.c_int(w), ctypes.c_int(h), ctypes.c_int(1), warp_matrix.ctypes.data_as(doublep),
            ctypes.c_int(nscales),ctypes.c_int(nparams),ctypes.c_int(robust), ctypes.c_double(lambda_value))
        
        # Translate into optical flow
        warp_matrix = warp_matrix.reshape(3, 3)
        X, Y = np.meshgrid(np.arange(w), np.arange(h))
        X, Y = X.flatten(), Y.flatten() 
        samplesloc = np.vstack(( X , Y, np.ones_like(X) ))
        newsamplesloc = np.linalg.inv(warp_matrix)@ samplesloc
        newsamplesloc[0,:]/=newsamplesloc[2,:]
        newsamplesloc[1,:]/=newsamplesloc[2,:]
        flow[:h*w] = newsamplesloc[0,:] - samplesloc[0,:]
        flow[h*w:2*h*w] = newsamplesloc[1,:] - samplesloc[1,:]

        return flow.reshape(2,h,w).transpose(1,2,0)

    def TVL1_flow(self, Im1, Im2):   
        """ Im1 and Im2 are iio RGB images.
        They will be translated to gray images inside this code.
        """   
        h, w = Im1.shape[:2]
        h1, w1 = Im2.shape[:2]
        assert h1==h and w1==w, "Both images Im1 and Im2 are supposed to share same size"

        I1 = np.zeros(h*w, dtype = ctypes.c_float)
        I2 = np.zeros(h*w, dtype = ctypes.c_float)
        flow = np.zeros(2*h*w, dtype = ctypes.c_float)

        if Im1.shape[2]==3:
            I1[:] = rgb2gray(Im1).flatten()[:]
            I2[:] = rgb2gray(Im2).flatten()[:]
        elif Im1.shape[2] == 4:
            I1[:] = np.mean(Im1, axis=2).flatten()[:]
            I2[:] = np.mean(Im2, axis=2).flatten()[:]
        elif Im1.shape[2] == 1:
            I1[:] = Im1.flatten()[:]
            I2[:] = Im2.flatten()[:]
        
        floatp = ctypes.POINTER(ctypes.c_float)
        self.libBridge.tvl1flow(I1.ctypes.data_as(floatp), I2.ctypes.data_as(floatp), flow.ctypes.data_as(floatp), ctypes.c_int(w), ctypes.c_int(h))
        # flow = np.resize(data, (int(h), int(w), 2))	
        return flow.reshape(2,h,w).transpose(1,2,0)
    
    def WarpWithBackwardFlow(self, Im1, Im2, flowtype='tvl1', noTorch=True, giveMeFlow=False):
        """ Im1 is warped into Im2 following the backward optical flow
        Im1, Im2 should have values around [0,255]. They can be floats!
        """
        assert Im1.dtype == Im2.dtype, 'Image types should be at least the same (%s,%s). Convert before calling this function.'%(Im1.dtype, Im2.dtype)
        if flowtype == 'tvl1':
            flow = self.TVL1_flow(Im2,Im1)
        elif flowtype == 'ica':
            flow = self.invcomp_flow(Im2,Im1)
        if giveMeFlow:
            if noTorch == True:
                return single_warp_noTorch(Im1, flow), flow
            else:
                return single_warp_Torch(Im1, flow), flow
        else:
            if noTorch==True:
                return single_warp_noTorch(Im1,flow)
            else:
                return single_warp_Torch(Im1,flow)

    def downsa(self, img, n=2, satype='v'):
        """ Downsample as in imscript
        """
        satype = ctypes.c_char(ord(satype))
        h, w, c = img.shape[:]

        W = int(w/n)
        H = int(h/n)
        Iout = np.zeros(H*W*c, dtype=ctypes.c_float)

        Iin = np.zeros(h*w*c, dtype=ctypes.c_float)
        Iin[:] = img.flatten()[:]        

        floatp = ctypes.POINTER(ctypes.c_float)
        self.libBridge.downsa(Iin.ctypes.data_as(floatp), Iout.ctypes.data_as(floatp),
                            ctypes.c_int(w), ctypes.c_int(h), ctypes.c_int(c), ctypes.c_int(n),
                             satype)
        return Iout.reshape(H,W,c)
    
    def upsa(self, img, n=2, uptype=2, offx=0, offy=0):
        """Upsample as in imscript
        """
        h, w, c = img.shape[:]

        W = n * w - n
        H = n * h - n
        Iout = np.zeros(H*W*c, dtype=ctypes.c_float)

        Iin = np.zeros(h*w*c, dtype=ctypes.c_float)
        Iin[:] = img.flatten()[:]

        floatp = ctypes.POINTER(ctypes.c_float)
        self.libBridge.upsa(Iin.ctypes.data_as(floatp), Iout.ctypes.data_as(floatp),
                              ctypes.c_int(w), ctypes.c_int(h), ctypes.c_int(c), ctypes.c_int(n),
                              ctypes.c_int(uptype), ctypes.c_float(offx), ctypes.c_float(offy) )
        return Iout.reshape(H, W, c)

    def blur(self, img, kernel='g', size="2", boundary_symmetric=True, boundary_periodic=False, boundary_zero=False):
        """ Blur as in imscript.
        Kernels:
	         'g': gaussian;
	         'l': laplace; 
	         'c': cauchy;  
	         'k': bicauchy;  
	         'q': logcauchy;  
	         'u': goodcauchy;  
	         'd': disk;    
	         's': square;  
	         'p': powerlaw2;  
	         'a': pareto;  
	         'i': invr;  
	         'r': land;  
	         'y': ynvr;  
	         'z': ilogr;  
	         't': r2logr;  
	         'o': logr;  
        """
        h, w, c = img.shape[:]

        Iin = np.zeros(h*w*c, dtype=ctypes.c_float)
        Iin[:] = img.flatten()[:]

        Iout = np.zeros(h*w*c, dtype=ctypes.c_float)

        Kparams = np.zeros(10, dtype=ctypes.c_char)
        Kid = np.zeros(1, dtype=ctypes.c_char)
        Kid[0] = kernel
        size = list(size)
        for i in range(len(size)):
            Kparams[i] = size[i]

        floatp = ctypes.POINTER(ctypes.c_float)
        charp = ctypes.POINTER(ctypes.c_char)
        self.libBridge.blur(Iin.ctypes.data_as(floatp), Iout.ctypes.data_as(floatp),
                            ctypes.c_int(w), ctypes.c_int(h), ctypes.c_int(c),
                            Kid.ctypes.data_as(
                                charp), Kparams.ctypes.data_as(charp),
                            ctypes.c_bool(boundary_symmetric), ctypes.c_bool(
                                boundary_periodic),
                            ctypes.c_bool(boundary_zero))
        return Iout.reshape(h,w,c)

    def warping_error(self, img, warped):
        """ Computes the warping error
        """
        img = self.upsa(self.blur(self.downsa(img)))
        warped = self.upsa(self.blur(self.downsa(warped)))
        return np.expand_dims(np.sum(np.fabs(img-warped), axis=2), 2)

    def compute_mask(self, img, warped, taus_size = 201, factor=5, bandwidth= 0.5):
        img = img.astype(np.float32)
        warped = warped.astype(np.float32)
        taus = np.linspace(0, 20, taus_size)

        k = self.warping_error(img, warped)
        mask_invalid_pixels = np.zeros(shape=np.shape(k))
        valid_k = k * (mask_invalid_pixels == 0)
        kde = stats.gaussian_kde(valid_k.flatten(), bandwidth)
        dd = kde(taus)

        #Automatic threshold
        imax = np.int(np.argmax(dd))
        mod = taus[imax]

        # med = np.median(k.flatten())
        scale_iqr = mod - np.percentile(k.flatten(), 10)
        tau = mod + factor*scale_iqr

        mtau = 1.*(k < tau)  # warping errors flagged with value 1
        mtau[mask_invalid_pixels > 0] = 2  # out-of-frame pixels with value 2
        output = np.zeros((mtau.shape[0]+2, mtau.shape[1]+2, 1))
        output[1:-1, 1:-1] = mtau.astype(np.uint8)
        return output
